﻿using System.Collections.Concurrent;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Xml.Serialization;
using ProtoBuf;

namespace Zekken
{
    public class ShapeCompiler
    {
        private static XmlSerializer xmlSerializer = new XmlSerializer(typeof(List<XSpellShape>));

        public static void CompileShapes()
        {
            if (!FileTools.PrepareDirectory(Directories.DEFAULT_SHAPES_DIR)) { return; }

            var files = GetXmlFilesFrom(Directories.DEFAULT_SHAPES_DIR)
                .ToList();

            foreach (var file in files) { CleanFile(file); }

            var shapes = files.SelectMany(LoadShapes).ToList();

            foreach (var shape in shapes)
            {
                shape.Flags = shape.Flags.Distinct().ToList();
            }

            foreach (var file in files) { File.Delete(file); }

            var purgedShapes = GetPurgedShapes(shapes).OrderBy(s => s.Id).ToList();
            Save(Directories.DEFAULT_XDB_PATH, purgedShapes);

            var compiledData = new ConcurrentDictionary<uint, SpellShape>();
            foreach (var shape in purgedShapes) { compiledData[shape.Id] = shape.GetShape(); }
            Save(compiledData);

            Logger.WriteMessage("Finished compiling {0} spell shapes.", purgedShapes.Count);
        }

        public ConcurrentDictionary<uint, SpellShape> Load()
        {
            return Load(Directories.DEFAULT_DATA_PATH);
        }

        public ConcurrentDictionary<uint, SpellShape> Load(string path)
        {
            if (!FileTools.PrepareLoad(path)) { return new ConcurrentDictionary<uint, SpellShape>(); }

            ConcurrentDictionary<uint, SpellShape> loadedShapes = null;
            using (var stream = new FileStream(path, FileMode.Open, FileAccess.Read))
            {
                try { loadedShapes = Serializer.Deserialize<ConcurrentDictionary<uint, SpellShape>>(stream); }
                catch { Logger.WriteError("Failed to load spell shape data from {0}.", path); }
            }

            return loadedShapes ?? new ConcurrentDictionary<uint, SpellShape>();
        }

        public static bool Save(ConcurrentDictionary<uint, SpellShape> data)
        {
            return Save(Directories.DEFAULT_DATA_PATH, data);
        }

        public static bool Save(string path, ConcurrentDictionary<uint, SpellShape> data)
        {
            if (!FileTools.PrepareSave(path)) { return false; }
            using (var stream = new FileStream(path, FileMode.Create, FileAccess.Write))
            {
                try { Serializer.Serialize(stream, data); }
                catch
                {
                    Logger.WriteError("Failed to save spell shape data at {0}", path);
                    return false;
                }
            }

            return true;
        }

        public static List<XSpellShape> LoadShapes(string path)
        {
            List<XSpellShape> shapes;

            using (var stream = new FileStream(path, FileMode.Open, FileAccess.Read))
            {
                try { shapes = (List<XSpellShape>)xmlSerializer.Deserialize(stream); }
                catch
                {
                    Logger.WriteError("Could not deserialize shapes from {0}.", path);
                    shapes = new List<XSpellShape>();
                }
            }

            return shapes;
        }

        public static void Save(string path, List<XSpellShape> shapes)
        {
            if (!FileTools.PrepareSave(path)) { return; }
            using (var stream = new FileStream(path, FileMode.Create, FileAccess.Write))
            {
                xmlSerializer.Serialize(stream, shapes);
            }
        }

        public static IEnumerable<string> GetXmlFilesFrom(string directory)
        {
            return Directory.EnumerateFiles(directory, "*.xml");
        }

        public static void Merge(XSpellShape a, XSpellShape b)
        {
            foreach (var flag in b.Flags) { a.Flags.Add(flag); }
        }

        public static List<XSpellShape> GetPurgedShapes(IEnumerable<XSpellShape> shapes)
        {
            var purgedShapes = new List<XSpellShape>();
            foreach (var shape in shapes)
            {
                var repeated = purgedShapes.FirstOrDefault(s => s.Id == shape.Id);
                if (repeated != null) { repeated.MergeWith(shape); }
                else { purgedShapes.Add(shape); }
            }

            return purgedShapes;
        }

        public static void CleanFile(string path)
        {
            try
            {
                var text = File.ReadAllText(path);
                text = text.Replace("AngleSmall", "WidthSmall");
                text = text.Replace("AngleMedium", "WidthMedium");
                text = text.Replace("AngleLarge", "WidthLarge");
                text = text.Replace("HeightSmall", "ReachSmall");
                text = text.Replace("HeightMedium", "ReachMedium");
                text = text.Replace("HeightLarge", "ReachLarge");
                text = text.Replace("RadiusSmall", "ReachSmall");
                text = text.Replace("RadiusMedium", "ReachMedium");
                text = text.Replace("RadiusLarge", "ReachLarge");
                text = text.Replace("<Flags>TargetedOnOther</Flags>", "TargetedOnRandom");
                File.WriteAllText(path, text);
            }
            catch { Logger.WriteError("Failed to clean file: {0}", path); }
        }
    }
}